#! /usr/bin/env python

import time
import random

class GnomesJob(object):
    @staticmethod
    def underpants(job):
        # Let's collect some underpants
        if random.random() < 0.005:
            raise Exception('Sometimes underpants are hard to get')
        else:
            # It takes time to collect underpants
            time.sleep(random.random() * 0.5)
            job['underpants'] = {
                'collected': random.randrange(0, 200)
            }
            job.complete('unknown')
    
    @staticmethod
    def unknown(job):
        # After all, it is the unknown stage
        if random.random() < 0.05:
            raise Exception('The most uncertain plans of mice and men...')
        else:
            # We'll scratch our heads a little bit about what to do here
            time.sleep(random.random() * 2.0)
            job['unknown'] = ['?'] * int(random.random() * 10)
            job.complete('profit')
    
    @staticmethod
    def profit(job):
        # How much profits did we get?!
        job['profit'] = '$%.2f' % (random.uniform(0.5, 1.5) * job['underpants']['collected'])
        job.complete()

if __name__ == '__main__':
    import qless
    from qless import gnomes
    client = qless.client()
    underpants = client.queue('underpants')
    for i in range(1000):
        underpants.put(gnomes.GnomesJob, {})



#! /usr/bin/env python

import simplejson as json

class Config(object):
    def __init__(self, client):
        self.client = client
    
    def __getattr__(self, attr):
        '''GetConfig(0, [option])
        ----------------------
        Get the current configuration value for that option, or if option is omitted,
        then get all the configuration values.'''
        if attr == 'all':
            return json.loads(self.client._config([], ['get']))
        raise AttributeError('qless.Config has no attribute %s' % attr)
    
    def __len__(self):
        return len(self.all)
    
    def __getitem__(self, option):
        return self.client._config([], ['get', option])
    
    def __setitem__(self, option, value):
        '''SetConfig(0, option, [value])
        -----------------------------
        Set the configuration value for the provided option. If `value` is omitted,
        then it will remove that configuration option.'''
        return self.client._config([], ['set', option, value])
    
    def __delitem__(self, option):
        return self.client._config([], ['unset', option])
    
    def __contains__(self, option):
        return dict.__contains__(self.all, option)
    
    def __iter__(self):
        return iter(self.all)
    
    def clear(self):
        for key in self.all.keys():
            self.client._config([], ['unset', key])
    
    def get(self, option, default=None):
        r = self[option]
        return ((r == None) and default) or r
    
    def items(self):
        return self.all.items()
    
    def keys(self):
        return self.all.keys()
    
    def pop(self, option, default=None):
        r = self[option]
        del self[option]
        return ((r == None) and default) or r
    
    def update(self, other=(), **kwargs):
        _kwargs = dict(kwargs)
        _kwargs.update(other)
        for key, value in _kwargs.items():
            self[key] = value
    
    def values(self):
        return self.all.values()



#! /usr/bin/env python

import os
import imp
import uuid
import time
import types
import traceback
from qless import logger
import simplejson as json

class BaseJob(object):
    # This is a dictionary of all the classes that we've seen, and
    # the last load time for each of them. We'll use this either for
    # the debug mode or the general mechanism
    _loaded = {}
    
    def __init__(self, client, **kwargs):
        self.client = client
        for att in ['data', 'jid', 'priority']:
            object.__setattr__(self, att, kwargs[att])
        
        object.__setattr__(self, 'klass_name', kwargs['klass'])
        object.__setattr__(self, 'queue_name', kwargs['queue'])
        # Because of how Lua parses JSON, empty tags comes through as {}
        object.__setattr__(self, 'tags'      , kwargs['tags'] or [])
    
    def __setattr__(self, key, value):
        if key == 'priority':
            return self.client._priority([], [self.jid, value]) and object.__setattr__(self, key, value)
        else:
            return object.__setattr__(self, key, value)
    
    def __getattr__(self, key):
        if key == 'queue':
            # An actual queue instance
            object.__setattr__(self, 'queue', self.client.queues[self.queue_name])
            return self.queue
        elif key == 'klass':
            # Get a reference to the provided klass
            object.__setattr__(self, 'klass', self._import(self.klass_name))
            return self.klass
        raise AttributeError('%s has no attribute %s' % (self.__class__.__module__ + '.' + self.__class__.__name__, key))
    
    @staticmethod
    def _import(klass):
        # 1) Get a reference to the module
        # 2) Check the file that module's imported from
        # 3) If that file's been updated, force a reload of that module, return it
        mod = __import__(klass.rpartition('.')[0])
        for m in klass.split('.')[1:-1]:
            mod = getattr(mod, m)
        
        # Alright, now check the file associated with it
        mtime = os.stat(mod.__file__).st_mtime
        BaseJob._loaded[klass] = min(BaseJob._loaded.get(klass, 1e20), time.time())
        if BaseJob._loaded[klass] < mtime:
            mod = reload(mod)
        
        return getattr(mod, klass.rpartition('.')[2])
    
    def cancel(self):
        '''Cancel a job. It will be deleted from the system, the thinking being that
        if you don't want to do any work on it, it shouldn't be in the queueing system.'''
        return self.client._cancel([], [self.jid])
    
    def tag(self, *tags):
        '''Tag a job with additional tags'''
        return self.client._tag([], ['add', self.jid, repr(time.time())] + list(tags))
    
    def untag(self, *tags):
        '''Remove tags from a job'''
        return self.client._tag([], ['remove', self.jid, repr(time.time())] + list(tags))

# The Job class
class Job(BaseJob):
    def __init__(self, client, **kwargs):
        BaseJob.__init__(self, client, **kwargs)
        self.client = client
        for att in ['state', 'tracked', 'failure', 'history', 'dependents', 'dependencies']:
            object.__setattr__(self, att, kwargs[att])
        
        # The reason we're using object.__setattr__ directly is because
        # we have __setattr__ defined for this class, and we're actually
        # just interested in setting these memebers directly
        object.__setattr__(self, 'expires_at'      , kwargs['expires'])
        object.__setattr__(self, 'original_retries', kwargs['retries'])
        object.__setattr__(self, 'retries_left'    , kwargs['remaining'])
        object.__setattr__(self, 'worker_name'     , kwargs['worker'])
        # Because of how Lua parses JSON, empty lists come through as {}
        object.__setattr__(self, 'dependents'      , kwargs['dependents'] or [])
        object.__setattr__(self, 'dependencies'    , kwargs['dependencies'] or [])
    
    def __getattr__(self, key):
        if key == 'ttl':
            # How long until this expires, in seconds
            return self.expires_at - time.time()
        return BaseJob.__getattr__(self, key)
    
    def __getitem__(self, key):
        return self.data.get(key)
    
    def __setitem__(self, key, value):
        self.data[key] = value
    
    def __repr__(self):
        return '<%s %s>' % (self.klass_name, self.jid)
    
    def process(self):
        '''Load the module containing your class, and run the appropriate method. For example,
        if this job was popped from the queue ``testing``, then this would invoke the ``testing``
        staticmethod of your class.'''
        method = getattr(self.klass, self.queue_name, getattr(self.klass, 'process', None))
        if method:
            if isinstance(method, types.FunctionType):
                try:
                    logger.info('Processing %s in %s' % (self.jid, self.queue_name))
                    method(self)
                    logger.info('Completed %s in %s' % (self.jid, self.queue_name))
                except Exception as e:
                    # Make error type based on exception type
                    logger.exception('Failed %s in %s: %s' % (self.jid, self.queue_name, repr(method)))
                    self.fail(self.queue_name + '-' + e.__class__.__name__, traceback.format_exc())
            else:
                # Or fail with a message to that effect
                logger.error('Failed %s in %s : %s is not static' % (self.jid, self.queue_name, repr(method)))
                self.fail(self.queue_name + '-method-type', repr(method) + ' is not static')
        else:
            # Or fail with a message to that effect
            logger.error('Failed %s : %s is missing a method "%s" or "process"' % (self.jid, self.klass_name, self.queue_name))
            self.fail(self.queue_name + '-method-missing', self.klass_name + ' is missing a method "' + self.queue_name + '" or "process"')
    
    def move(self, queue, delay=0, depends=None):
        '''Move this job out of its existing state and into another queue. If a worker
        has been given this job, then that worker's attempts to heartbeat that job will
        fail. Like ``Queue.put``, this accepts a delay, and dependencies'''
        logger.info('Moving %s to %s from %s' % (self.jid, queue, self.queue_name))
        return self.client._put([queue], [
            self.jid,
            self.klass_name,
            json.dumps(self.data),
            repr(time.time()),
            delay,
            'depends', json.dumps(depends or [])
        ])
    
    def complete(self, next=None, delay=None, depends=None):
        '''Turn this job in as complete, optionally advancing it to another queue. Like
        ``Queue.put`` and ``move``, it accepts a delay, and dependencies'''
        if next:
            logger.info('Advancing %s to %s from %s' % (self.jid, next, self.queue_name))
            return self.client._complete([], [self.jid, self.client.worker_name, self.queue_name,
                repr(time.time()), json.dumps(self.data), 'next', next, 'delay', delay or 0,
                'depends', json.dumps(depends or [])]) or False
        else:
            logger.info('Completing %s' % self.jid)
            return self.client._complete([], [self.jid, self.client.worker_name, self.queue_name,
                repr(time.time()), json.dumps(self.data)]) or False
    
    def heartbeat(self):
        '''Renew the heartbeat, if possible, and optionally update the job's user data.'''
        self.expires_at = float(self.client._heartbeat([], [self.jid, self.client.worker_name, repr(time.time()), json.dumps(self.data)]) or 0)
        return self.expires_at
    
    def fail(self, group, message):
        '''Mark the particular job as failed, with the provided type, and a more specific
        message. By `type`, we mean some phrase that might be one of several categorical
        modes of failure. The `message` is something more job-specific, like perhaps
        a traceback.
        
        This method should __not__ be used to note that a job has been dropped or has 
        failed in a transient way. This method __should__ be used to note that a job has
        something really wrong with it that must be remedied.
        
        The motivation behind the `type` is so that similar errors can be grouped together.
        Optionally, updated data can be provided for the job. A job in any state can be
        marked as failed. If it has been given to a worker as a job, then its subsequent
        requests to heartbeat or complete that job will fail. Failed jobs are kept until
        they are canceled or completed. __Returns__ the id of the failed job if successful,
        or `False` on failure.'''
        logger.warn('Failing %s (%s): %s' % (self.jid, group, message))
        return self.client._fail([], [self.jid, self.client.worker_name, group, message, repr(time.time()), json.dumps(self.data)]) or False
    
    def track(self):
        '''Begin tracking this job'''
        return self.client._track([], ['track', self.jid, repr(time.time())])
    
    def untrack(self):
        '''Stop tracking this job'''
        return self.client._track([], ['untrack', self.jid, repr(time.time())])
    
    def retry(self, delay=0):
        '''Retry this job in a little bit, in the same queue. This is meant for the times
        when you detect a transient failure yourself'''
        result = self.client._retry([], [self.jid, self.queue_name, self.worker_name, repr(time.time()), delay])
        if result == None:
            return False
        return result
    
    def depend(self, *args):
        '''If and only if a job already has other dependencies, this will add more jids
        to the list of this job's dependencies.'''
        return self.client._depends([], [self.jid, 'on'] + list(args)) or False
    
    def undepend(self, *args, **kwargs):
        '''Remove specific (or all) job dependencies from this job::
            
            job.remove(jid1, jid2)
            job.remove(all=True)'''
        if kwargs.get('all', False):
            return self.client._depends([], [self.jid, 'off', 'all']) or False
        else:
            return self.client._depends([], [self.jid, 'off'] + list(args)) or False

class RecurringJob(BaseJob):
    def __init__(self, client, **kwargs):
        for att in ['data', 'jid', 'priority', 'tags', 'retries', 'interval', 'count']:
            object.__setattr__(self, att, kwargs[att])
        
        object.__setattr__(self, 'client'    , client         )
        object.__setattr__(self, 'klass_name', kwargs['klass'])
        object.__setattr__(self, 'queue_name', kwargs['queue'])
        object.__setattr__(self, 'tags'      , self.tags or [])
    
    def __setattr__(self, key, value):
        if key in ('priority', 'retries', 'interval'):
            return self.client._recur([], ['update', self.jid, key, value]) and object.__setattr__(self, key, value)
        if key == 'data':
            return self.client._recur([], ['update', self.jid, key, json.dumps(value)]) and object.__setattr__(self, 'data', value)
        if key == 'klass':
            name = value.__module__ + '.' + value.__name__
            return self.client._recur([], ['update', self.jid, 'klass', name]) and object.__setattr__(self, 'klass_name', name) and object.__setattr__(self, 'klass', value)
        return object.__setattr__(self, key, value)
    
    def __getattr__(self, key):
        if key == 'next':
            # The time (seconds since epoch) until the next time it's run
            return self.client.redis.zscore('ql:q:' + self.queue_name + '-recur', self.jid)
        return BaseJob.__getattr__(self, key)
    
    def move(self, queue):
        '''Make this recurring job attached to another queue'''
        return self.client._recur([], ['update', self.jid, 'queue', queue])
    
    def cancel(self):
        '''Cancel all future recurring jobs'''
        self.client._recur([], ['off', self.jid])
    
    def tag(self, *tags):
        return self.client._recur([], ['tag', self.jid] + list(tags))
    
    def untag(self, *tags):
        return self.client._recur([], ['untag', self.jid] + list(tags))


#! /usr/bin/env python

import time
import uuid
from job import Job
import simplejson as json

class Jobs(object):
    def __init__(self, name, client):
        self.name   = name
        self.client = client
    
    def running(self, offset=0, count=25):
        return self.client._jobs([], ['running', repr(time.time()), self.name, offset, count])
    
    def stalled(self, offset=0, count=25):
        return self.client._jobs([], ['stalled', repr(time.time()), self.name, offset, count])
    
    def scheduled(self, offset=0, count=25):
        return self.client._jobs([], ['scheduled', repr(time.time()), self.name, offset, count])
    
    def depends(self, offset=0, count=25):
        return self.client._jobs([], ['depends', repr(time.time()), self.name, offset, count])
    
    def recurring(self, offset=0, count=25):
        return self.client._jobs([], ['recurring', repr(time.time()), self.name, offset, count])

# The Queue class
class Queue(object):
    def __init__(self, name, client, worker_name):
        self.name        = name
        self.client      = client
        self.worker_name = worker_name
        self._hb         = 60
    
    def __getattr__(self, key):
        if key == 'jobs':
            self.jobs = Jobs(self.name, self.client)
            return self.jobs
        if key == 'counts':
            return json.loads(self.client._queues([], [time.time(), self.name]))
        if key == 'heartbeat':
            a = client.config.all
            return int(a.get(self.name + '-heartbeat', a.get('heartbeat', 60)))
        raise AttributeError('qless.Queue has no attribute %s' % key)
    
    def __setattr__(self, key, value):
        if key == 'heartbeat':
            self.client.config[self.name + '-heartbeat'] = value
        else:
            object.__setattr__(self, key, value)
    
    def put(self, klass, data, priority=None, tags=None, delay=None, retries=None, jid=None, depends=None):
        '''Either create a new job in the provided queue with the provided attributes,
        or move that job into that queue. If the job is being serviced by a worker,
        subsequent attempts by that worker to either `heartbeat` or `complete` the
        job should fail and return `false`.
            
        The `priority` argument should be negative to be run sooner rather than 
        later, and positive if it's less important. The `tags` argument should be
        a JSON array of the tags associated with the instance and the `valid after`
        argument should be in how many seconds the instance should be considered 
        actionable.'''
        return self.client._put([self.name], [
            jid or uuid.uuid4().hex,
            klass.__module__ + '.' + klass.__name__,
            json.dumps(data),
            repr(time.time()),
            delay or 0,
            'priority', priority or 0,
            'tags', json.dumps(tags or []),
            'retries', retries or 5,
            'depends', json.dumps(depends or [])
        ])
    
    def recur(self, klass, data, interval, offset=0, priority=None, tags=None, retries=None, jid=None):
        return self.client._recur([], ['on', self.name,
            jid or uuid.uuid4().hex,
            klass.__module__ + '.' + klass.__name__,
            json.dumps(data),
            repr(time.time()),
            'interval', interval, offset,
            'priority', priority or 0,
            'tags', json.dumps(tags or []),
            'retries', retries or 5
        ])
    
    def pop(self, count=None):
        '''Passing in the queue from which to pull items, the current time, when the locks
        for these returned items should expire, and the number of items to be popped
        off.'''
        results = [Job(self.client, **json.loads(j)) for j in self.client._pop([self.name], [self.worker_name, count or 1, repr(time.time())])]
        if count == None:
            return (len(results) and results[0]) or None
        return results
    
    def peek(self, count=None):
        '''Similar to the pop command, except that it merely peeks at the next items'''
        results = [Job(self.client, **json.loads(r)) for r in self.client._peek([self.name], [count or 1, repr(time.time())])]
        if count == None:
            return (len(results) and results[0]) or None
        return results
    
    def stats(self, date=None):
        '''Return the current statistics for a given queue on a given date. The results 
        are returned are a JSON blob::
        
            {
                'total'    : ...,
                'mean'     : ...,
                'variance' : ...,
                'histogram': [
                    ...
                ]
            }
        
        The histogram's data points are at the second resolution for the first minute,
        the minute resolution for the first hour, the 15-minute resolution for the first
        day, the hour resolution for the first 3 days, and then at the day resolution
        from there on out. The `histogram` key is a list of those values.'''
        return json.loads(self.client._stats([], [self.name, date or repr(time.time())]))
    
    def __len__(self):
        with self.client.redis.pipeline() as p:
            o = p.zcard('ql:q:' + self.name + '-locks')
            o = p.zcard('ql:q:' + self.name + '-work')
            o = p.zcard('ql:q:' + self.name + '-scheduled')
            return sum(p.execute())



#! /usr/bin/env python

import os
import math
import time
import qless
import shutil
import psutil
import signal
import logging
from qless import logger

# This meta-information about the worker you're running on
meta = {
    'worker_id': 0,
    'sandbox'  : '/dev/null'
}

try:
    from setproctitle import setproctitle
except ImportError:
    def setproctitle(title):
        pass

class Worker(object):
    def __init__(self, queues, host='localhost', workers=None, interval=60, workdir='.', resume=False, **kwargs):
        _host, s, _port = host.partition(':')
        _port           = int(_port or 6379)
        self.host       = _host
        self.port       = _port
        self.client     = qless.client(self.host, self.port)
        self.count      = workers or psutil.NUM_CPUS
        self.interval   = interval
        self.queues     = queues
        self.resume     = resume
        self.sandboxes  = {}
        # This is for filesystem sandboxing. Each worker has
        # a directory associated with it, which it should make
        # sure is the working directory in which it runs each
        # of the jobs. It should also ensure that the directory
        # exists, and clobbers files before each run, and after.
        self.workdir    = os.path.abspath(workdir)
        self.sandbox    = self.workdir
        # I'm the parent, so I have a negative worker id
        self.worker_id  = -1
        self.master     = True
        # These are the job ids that I should get to first, before
        # picking off other jobs
        self.jids       = []
    
    def run(self):
        # If this worker is meant to be resumable, then we should find out
        # what jobs this worker was working on beforehand.
        if self.resume:
            jids_to_resume = self.client.workers[self.client.worker_name]['jobs']
        else:
            jids_to_resume = []
        
        for i in range(self.count):
            slot = {
                'worker_id': i,
                'sandbox'  : os.path.join(self.workdir, 'qless-py-workers', 'sandbox-%i' % i)
            }
            cpid = os.fork()
            if cpid:
                logger.info('Spawned worker %i' % cpid)
                self.sandboxes[cpid] = slot
            else:
                # Set the value of the metadata so that jobs can detect
                # what worker they're running on
                import qless.worker
                qless.worker.meta = slot
                # Make note that we're not the master, and then save our
                # sandbox and worker id for reference
                self.master    = False
                self.sandbox   = slot['sandbox']
                self.worker_id = slot['worker_id']
                # Also, we should take our share of the jobs that we want
                # to resume, if any.
                start = (i     * len(jids_to_resume)) / self.count
                end   = ((i+1) * len(jids_to_resume)) / self.count
                self.jids      = jids_to_resume[start:end]
                return self.work()
        
        while self.master:
            try:
                pid, status = os.wait()
                logger.warn('Worker %i died with status %i from signal %i' % (pid, status >> 8, status & 0xff))
                slot = self.sandboxes.pop(pid)
                cpid = os.fork()
                if cpid:
                    logger.info('Spawned replacement worker %i' % cpid)
                    self.sandboxes[cpid] = slot
                else:
                    # Set the value of the metadata so that jobs can detect
                    # what worker they're running on
                    import qless.worker
                    qless.worker.meta = slot
                    # Make note that we're not the master, and then save our
                    # sandbox and worker id for reference
                    self.master    = False
                    self.sandbox   = slot['sandbox']
                    self.worker_id = slot['worker_id']
                    # NOTE: In the case that the worker died, we're going to
                    # assume that something about the job(s) it was working 
                    # made the worker exit, and so we're going to ignore any
                    # jobs that we might have been working on. It's also 
                    # significantly more difficult than the above problem of
                    # simply distributing work to /new/ workers, rather than
                    # a respawned worker.
                    return self.work()
            except KeyboardInterrupt:
                break
        
        if self.master:
            self.stop()
    
    def clean(self):
        # This cleans the sandbox -- changing the working directory to it,
        # as well as clearing out any files that might be in there
        # Make sure we're running in our sandbox
        os.chdir(self.sandbox)
        # And that it's clear of any files
        for p in os.listdir(self.sandbox):
            p = os.path.join(self.sandbox, p)
            if os.path.isdir(p):
                logger.info('Removing tree %s...' % p)
                shutil.rmtree(p)
            else:
                logger.info('Removing file %s...' % p)
                os.remove(p)
    
    def setproctitle(self, message):
        base = 'qless-py-worker [%s] ' % ','.join(q.name for q in self.queues)
        setproctitle(base + message)
    
    def work(self):
        # We should probably open up our own redis client
        self.client = qless.client(self.host, self.port)
        self.queues = [self.client.queues[q] for q in self.queues]
        
        if not os.path.isdir(self.sandbox):
            os.makedirs(self.sandbox)
        self.clean()
        # First things first, we should clear out any jobs that
        # we're responsible for off-hand
        while len(self.jids):
            try:
                job = self.client.jobs[self.jids.pop(0)]
                # If we still have access to it, then we should process it
                if job.heartbeat():
                    logger.info('Resuming %s' % job.jid)
                    self.setproctitle('Working %s (%s)' % (job.jid, job.klass_name))
                    job.process()
                    self.clean()
                else:
                    logger.warn('Lost heart on would-be resumed job %s' % job.jid)
            except KeyboardInterrupt:
                return
        
        while True:
            try:
                seen = False
                for queue in self.queues:
                    job = queue.pop()
                    if job:
                        seen = True
                        self.setproctitle('Working %s (%s)' % (job.jid, job.klass_name))
                        job.process()
                        self.clean()
                
                if not seen:
                    self.setproctitle('sleeping...')
                    logger.debug('Sleeping for %fs' % self.interval)
                    time.sleep(self.interval)
            except KeyboardInterrupt:
                return
    
    def stop(self):
        # Stop all the workers, and then wait for them
        for cpid in self.sandboxes.keys():
            logger.warn('Stopping %i...' % cpid)
            os.kill(cpid, signal.SIGINT)
        
        while True:
            try:
                pid, status = os.wait()
                self.sandboxes.pop(pid, None)
                logger.warn('Worker %i stopped.' % cpid)
            except OSError:
                break
        
        for cpid in self.sandboxes.keys():
            logger.warn('Could not wait for %i' % cpid)

    # QUIT - Wait for child to finish processing then exit
    # TERM / INT - Immediately kill child then exit
    # USR1 - Immediately kill child but don't exit
    # USR2 - Don't start to process any new jobs
    # CONT - Start to process new jobs again after a USR2



#! /usr/bin/env python

import os
import qless
from qless import worker, logger

class Worker(worker.Worker):
    def __init__(self, *args, **kwargs):
        self.pool_size = kwargs.pop('pool_size', 10)
        worker.Worker.__init__(self, *args, **kwargs)
    
    def work(self):
        # We should probably open up our own redis client
        self.client = qless.client(self.host, self.port)
        self.queues = [self.client.queues[q] for q in self.queues]
        
        if not os.path.isdir(self.sandbox):
            os.makedirs(self.sandbox)
        
        from gevent.pool import Pool
        from gevent import sleep, Greenlet
        pool = Pool(self.pool_size)
        while True:
            try:
                seen = False
                for queue in self.queues:
                    # Wait until a greenlet is available
                    pool.wait_available()
                    job = queue.pop()
                    if job:
                        # For whatever reason, doing imports within a greenlet
                        # (there's one implicitly invoked in job.process), was
                        # throwing exceptions. The relatively ghetto way to get
                        # around this is to force the import to happen before
                        # the greenlet is spawned.
                        _module = job.klass
                        seen = True
                        pool.start(Greenlet(job.process))
                
                if not seen:
                    logger.debug('Sleeping for %fs' % self.interval)
                    sleep(self.interval)
            except KeyboardInterrupt:
                return



#! /usr/bin/env python

import pkgutil
import logging
logger = logging.getLogger('qless')

# The Lua class
class lua(object):
    def __init__(self, name, r):
        self.name  = name
        self.redis = r
        self.sha   = None
    
    def reload(self):
        data = pkgutil.get_data('qless', 'qless-core/' + self.name + '.lua')
        self.sha = self.redis.execute_command('script', 'load', data)
        logger.debug('Loaded script %s (%s)' % (self.name, repr(self.sha)))
    
    def __call__(self, keys, args):
        if self.sha == None:
            self.reload()
        try:
            return self.redis.execute_command('evalsha', self.sha, len(keys), *(keys + args))
        except Exception as e:
            self.reload()
            return self.redis.execute_command('evalsha', self.sha, len(keys), *(keys + args))



#! /usr/bin/env python

import os
import time
import redis
import logging
import simplejson as json

logger = logging.getLogger('qless')
formatter = logging.Formatter('%(asctime)s | PID %(process)d | [%(levelname)s] %(message)s')
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.FATAL)

# A decorator to specify a bunch of exceptions that should be caught
# and the job retried. It turns out this comes up with relative frequency
def retry(*exceptions):
    def decorator(f):
        def _f(job):
            try:
                f(job)
            except tuple(exceptions):
                job.retry()
        return _f
    return decorator

class Jobs(object):
    def __init__(self, client):
        self.client = client
    
    def complete(self, offset=0, count=25):
        '''Return the paginated jids of complete jobs'''
        return self.client._jobs([], ['complete', offset, count])
    
    def tracked(self):
        '''Return an array of job objects that are being tracked'''
        results = json.loads(self.client._track([], []))
        results['jobs'] = [Job(self, **j) for j in results['jobs']]
        return results
    
    def tagged(self, tag, offset=0, count=25):
        '''Return the paginated jids of jobs tagged with a tag'''
        return json.loads(self.client._tag([], ['get', tag, offset, count]))
    
    def failed(self, group=None, start=0, limit=25):
        '''If no group is provided, this returns a JSON blob of the counts of the various
        types of failures known. If a type is provided, returns paginated job objects 
        affected by that kind of failure.'''
        if not group:
            return json.loads(self.client._failed([], []))
        else:
            results = json.loads(self.client._failed([], [group, start, limit]))
            results['jobs'] = [Job(self.client, **j) for j in results['jobs']]
            return results
    
    def __getitem__(self, id):
        '''Get a job object corresponding to that jid, or ``None`` if it doesn't exist'''
        results = self.client._get([], [id])
        if not results:
            results = self.client._recur([], ['get', id])
            if not results:
                return None
            return RecurringJob(self.client, **json.loads(results))
        return Job(self.client, **json.loads(results))

class Workers(object):
    def __init__(self, client):
        self.client = client
    
    def __getattr__(self, attr):
        '''What workers are workers, and how many jobs are they running'''
        if attr == 'counts':
            return json.loads(self.client._workers([], [time.time()]))
        raise AttributeError('qless.Workers has no attribute %s' % attr)
    
    def __getitem__(self, worker_name):
        '''Which jobs does a particular worker have running'''
        result = json.loads(self.client._workers([], [time.time(), worker_name]))
        result['jobs']    = result['jobs'] or []
        result['stalled'] = result['stalled'] or []
        return result

class Queues(object):
    def __init__(self, client):
        self.client = client
    
    def __getattr__(self, attr):
        '''What queues are there, and how many jobs do they have running, waiting, 
        scheduled, etc.'''
        if attr == 'counts':
            return json.loads(self.client._queues([], [time.time()]))
        raise AttributeError('qless.Queues has no attribute %s' % attr)
    
    def __getitem__(self, queue_name):
        '''Get a queue object associated with the provided queue name'''
        return Queue(queue_name, self.client, self.client.worker_name)

class Events(object):
    def __init__(self, client):
        self.pubsub    = client.redis.pubsub()
        self.callbacks = dict(
            (k, None) for k in ('canceled', 'completed', 'failed', 'popped', 'stalled', 'put', 'track', 'untrack')
        )
        r = [self.pubsub.subscribe(k) for k in self.callbacks.keys()]
    
    def next(self):
        message = self.pubsub.listen().next()
        f = (message and message['type'] == 'message' and self.callbacks.get(message['channel']))
        if f:
            f(message['data'])
    
    def listen(self):
        try:
            while True:
                self.next()
        except redis.ConnectionError:
            return
    
    def on(self, evt, f):
        if evt not in self.callbacks:
            raise NotImplementedError('callback "%s"' % evt)
        else:
            self.callbacks[evt] = f
    
    def off(self, evt):
        return self.callbacks.pop(evt, None)

class client(object):
    def __init__(self, host='localhost', port=6379, hostname = None, **kwargs):
        import os
        import socket
        # This is our unique idenitifier as a worker
        self.worker_name = hostname or socket.gethostname()
        # This is just the redis instance we're connected to
        # conceivably someone might want to work with multiple
        # instances simultaneously.
        self.redis   = redis.Redis(host, port, **kwargs)
        self.events  = Events(self)
        self.config  = Config(self)
        self.jobs    = Jobs(self)
        self.workers = Workers(self)
        self.queues  = Queues(self)
        # Client's lua scripts
        for cmd in [
            'cancel', 'config', 'complete', 'depends', 'fail', 'failed', 'get', 'heartbeat', 'jobs', 'peek',
            'pop', 'priority', 'put', 'queues', 'recur', 'retry', 'stats', 'tag', 'track', 'workers']:
            setattr(self, '_%s' % cmd, lua(cmd, self.redis))
    
    def track(self, jid):
        '''Begin tracking this job'''
        return self._track([], ['track', jid, repr(time.time())])
    
    def untrack(self, jid):
        '''Stop tracking this job'''
        return self._track([], ['untrack', jid, repr(time.time())])
    
    def tags(self, offset=0, count=100):
        '''The most common tags among jobs'''
        return json.loads(self._tag([], ['top', offset, count]))
    
    def event(self):
        '''Listen for a single event'''
        pass
    
    def events(self, *args, **kwargs):
        '''Listen indefinitely for all events'''
        while True:
            self.event(*args, **kwargs)

from lua import lua
from job import Job, RecurringJob
from queue import Queue
from config import Config